"""
Representation of a collection of Doorstop items.
"""

import os
import logging

import yaml

from doorstop.core.item import Item
from doorstop.common import DoorstopError


class Document(object):
    """Represents a document containing an outline of items."""

    CONFIG = '.doorstop.yml'
    SKIP = '.doorstop.skip'  # indicates this document should be skipped
    DEFAULT_PREFIX = 'REQ'
    DEFAULT_PARENT = None  # which indicates this is the root document
    DEFAULT_DIGITS = 3

    def __init__(self, path, root=os.getcwd(),
                 _prefix=None, _parent=None, _digits=None):
        """Load a Document from an exiting directory.

        Internally, this constructor is also used to initialize new
        documents by providing default properites.

        @param path: path to Document directory
        @param root: path to root of project
        """
        # Check document's directory
        if not os.path.isfile(os.path.join(path, Document.CONFIG)):
            relpath = os.path.relpath(path, root)
            msg = "no {} in {}".format(Document.CONFIG, relpath)
            raise DoorstopError(msg)
        # Initialize Document
        self.path = path
        self.root = root
        self.prefix = _prefix or Document.DEFAULT_PREFIX
        self.parent = _parent or Document.DEFAULT_PARENT
        self.digits = _digits or Document.DEFAULT_DIGITS
        self.load()
        self.save()
        # Mark if skippable
        self.skip = os.path.isfile(os.path.join(self.path, Document.SKIP))

    def __repr__(self):
        return "Document({})".format(repr(self.path))

    def __str__(self):
        relpath = os.path.relpath(self.path, self.root)
        return "{} (@{}{})".format(self.prefix, os.sep, relpath)

    def __iter__(self):
        for filename in os.listdir(self.path):
            path = os.path.join(self.path, filename)
            try:
                yield Item(path)
            except DoorstopError as error:
                logging.debug(error)

    def __eq__(self, other):
        return isinstance(other, Document) and self.path == other.path

    def __ne__(self, other):
        return not self == other

    @property
    def config(self):
        """Get the path to the document's configuration."""
        return os.path.join(self.path, Document.CONFIG)

    @staticmethod
    def new(path, root, prefix, parent=None, digits=None):
        """Create a new Document.

        @param path: path to directory for the new document
        @param root: path to root of the project
        @param prefix: prefix for the new document
        @param parent: parent ID for the new document
        @param digits: number of digits for the new document

        @raise DoorstopError: if the document already exists
        """
        config = os.path.join(path, Document.CONFIG)
        # Check for an existing document
        if os.path.exists(config):
            raise DoorstopError("document already exists: {}".format(path))
        # Create the document directory
        Document._new(path, config)
        # Return the new document
        return Document(path, root=root,
                        _prefix=prefix, _parent=parent, _digits=digits)

    @staticmethod
    def _new(path, config):  # pragma: no cover, integration test
        """Create a new document directory.

        @param path: path to new document directory
        @param config: path to new document config file
        """
        if not os.path.exists(path):
            os.makedirs(path)
        with open(config, 'w'):
            pass  # just touch the file

    def load(self):
        """Load the document's properties from a file."""
        logging.debug("loading {}...".format(repr(self)))
        text = self._read()
        data = yaml.load(text)
        if data:
            settings = data.get('settings', {})
            if settings:
                self.prefix = settings.get('prefix', self.prefix)
                self.parent = settings.get('parent', self.parent)
                self.digits = settings.get('digits', self.digits)

    def _read(self):  # pragma: no cover, integration test
        """Read text from the file."""
        path = self.config
        if not os.path.exists(path):
            logging.debug("document does not exist yet: {}".format(path))
            return ""
        with open(path, 'rb') as infile:
            return infile.read().decode('UTF-8')

    def save(self):
        """Save the document's properties to a file."""
        logging.debug("saving {}...".format(repr(self)))
        settings = {'prefix': self.prefix,
                    'digits': self.digits}
        if self.parent:
            settings['parent'] = self.parent
        data = {'settings': settings}
        text = yaml.dump(data, default_flow_style=False)
        self._write(text)

    def _write(self, text):  # pragma: no cover, integration test
        """Write text to the file."""
        path = self.config
        with open(path, 'wb') as outfile:
            outfile.write(bytes(text, 'UTF-8'))

    @property
    def items(self):
        """Get an ordered list of items in the document."""
        return sorted(item for item in self)

    def add(self):
        """Create a new item for the document and return it."""
        number = self.maximum + 1
        logging.debug("next number: {}".format(number))
        try:
            last = self.items[-1]
        except IndexError:
            level = None
        else:
            level = last.level[:-1] + (last.level[-1] + 1,)
        logging.debug("next level: {}".format(level))
        return Item.new(self.path, self.root, self.prefix, self.digits,
                        number, level)

    @property
    def maximum(self):
        """Return the highest item number in the document."""
        try:
            return max(item.number for item in self)
        except ValueError:
            return 0

    def check(self, tree=None, ignored=None):
        """Confirm the document is valid.

        @param tree: tree to validate the document
        @param ignored: function to determine if a path should be skipped

        @return: indication that document is valid
        """
        logging.info("checking document {}...".format(self))
        items = list(self)
        # Check for items
        if not items:
            logging.warning("no items: {}".format(self))
        # Check each item
        for item in items:
            item.check(document=self, tree=tree, ignored=ignored)
        # Document is valid
        return True



#!/usr/bin/env python

"""
Unit tests for the doorstop.core.item module.
"""

import unittest
from unittest.mock import patch, Mock

import os
import logging

from doorstop.common import DoorstopError
from doorstop.core.item import Item

from doorstop.core.test import ENV, REASON, FILES, EMPTY


class MockItem(Item):
    """Item class with mock read/write methods."""

    @patch('os.path.isfile', Mock(return_value=True))
    def __init__(self, *args, **kwargs):
        self._file = kwargs.pop('_file', "")  # file system mock
        super().__init__(*args, **kwargs)
        self._read = Mock(side_effect=self._mock_read)
        self._write = Mock(side_effect=self._mock_write)

    def _mock_read(self):
        """Mock read method."""
        text = self._file
        logging.debug("mock read: {0}".format(repr(text)))
        return text

    def _mock_write(self, text):
        """Mock write method"""
        logging.debug("mock write: {0}".format(repr(text)))
        self._file = text

    _new = Mock()


class TestItem(unittest.TestCase):  # pylint: disable=R0904
    """Unit tests for the Item class."""  # pylint: disable=C0103,W0212

    def setUp(self):
        self.path = os.path.join('path', 'to', 'RQ001.yml')
        self.item = MockItem(self.path, _file=("links: []\n"
                                               "ref: ''\n"
                                               "text: ''\n"
                                               "level: 1.1.1"))

    def test_load_empty(self):
        """Verify loading calls read."""
        self.item.load()
        self.item._read.assert_called_once_with()
        self.assertEqual('', self.item._text)
        self.assertEqual(set(), self.item._links)
        self.assertEqual((1, 1, 1), self.item._level)
        self.item.check()

    def test_load_error(self):
        """Verify an exception is raised with invalid YAML."""
        self.item._file = "markdown: **Document and Item Creation**"
        self.assertRaises(DoorstopError, self.item.load)

    def test_save_empty(self):
        """Verify saving calls write."""
        self.item.save()
        text = ("level: 1\n"
                "links: []\n"
                "ref: ''\n"
                "text: ''\n")
        self.item._write.assert_called_once_with(text)

    def test_str(self):
        """Verify an item can be printed."""
        text = "RQ001 (@{}{})".format(os.sep, self.path)
        self.assertEqual(text, str(self.item))

    def test_ne(self):
        """Verify item non-equality is correct."""
        self.assertNotEqual(self.item, None)

    def test_lt(self):
        """Verify items can be compared."""
        item0 = MockItem('path/to/RQ002.yml', _level=(1, 1))
        item1 = self.item
        item2 = MockItem('path/to/RQ003.yml', _level=(1, 1, 2))
        self.assertLess(item0, item1)
        self.assertLess(item1, item2)
        self.assertGreater(item2, item0)

    def test_id(self):
        """Verify an item's ID can be read but not set."""
        self.assertEqual('RQ001', self.item.id)
        self.assertRaises(AttributeError, setattr, self.item, 'id', 'RQ002')

    def test_prefix(self):
        """Verify an item's prefix can be read but not set."""
        self.assertEqual('RQ', self.item.prefix)
        self.assertRaises(AttributeError, setattr, self.item, 'prefix', 'REQ')

    def test_number(self):
        """Verify an item's number can be read but not set."""
        self.assertEqual(1, self.item.number)
        self.assertRaises(AttributeError, setattr, self.item, 'number', 2)

    def test_level(self):
        """Verify an item's level can be set and read."""
        self.item.level = (1, 2, 3)
        self.assertIn("level: 1.2.3\n", self.item._write.call_args[0][0])
        self.assertEqual((1, 2, 3), self.item.level)

    def test_heading(self):
        """Verify the heading can be read from the item's level."""
        self.item.level = (1,)
        self.assertEqual(1, self.item.heading)
        self.item.level = (1, 0)
        self.assertEqual(1, self.item.heading)
        self.item.level = (2, 0, 1)
        self.assertEqual(3, self.item.heading)
        self.item.level = (2, 0, 1, 1, 0, 0)
        self.assertEqual(4, self.item.heading)

    def test_level_from_text(self):
        """Verify an item's level can be set from text and read."""
        self.item.level = "4.2.0 "
        self.assertIn("level: 4.2.0\n", self.item._write.call_args[0][0])
        self.assertEqual((4, 2, 0), self.item.level)

    def test_level_from_float(self):
        """Verify an item's level can be set from a float and read."""
        self.item.level = 4.2
        self.assertIn("level: 4.2\n", self.item._write.call_args[0][0])
        self.assertEqual((4, 2), self.item.level)

    def test_level_from_int(self):
        """Verify an item's level can be set from a int and read."""
        self.item.level = 42
        self.assertIn("level: 42\n", self.item._write.call_args[0][0])
        self.assertEqual((42,), self.item.level)

    def test_text(self):
        """Verify an item's text can be set and read."""
        self.item.text = "test "
        self.assertIn("text: |-\n  test\n", self.item._write.call_args[0][0])
        self.assertEqual("test", self.item.text)

    def test_text_sentences(self):
        """Verify newlines separate sentences in an item's text."""
        self.item.text = ("A sentence. Another sentence! Hello? Hi.\n"
                          "A new line. And another sentece.")
        expected = ("A sentence.\n"
                    "Another sentence!\n"
                    "Hello?\n"
                    "Hi.\n"
                    "A new line.\n"
                    "And another sentece.")
        self.assertEqual(expected, self.item.text)

    def test_ref(self):
        """Verify an item's reference can be set and read."""
        self.item.ref = "abc123"
        self.assertIn("ref: abc123\n", self.item._write.call_args[0][0])
        self.assertEqual("abc123", self.item.ref)

    def test_find_ref_error(self):
        """Verify an error occurs when no external reference found."""
        self.item.ref = "not found".replace(' ', '')  # avoids self match
        self.assertRaises(DoorstopError, self.item.find_ref, root=EMPTY)

    @unittest.skipUnless(os.getenv(ENV), REASON)
    def test_find_ref_error_long(self):
        """Verify an error occurs when no external reference found (long)."""
        self.item.ref = "not found".replace(' ', '')  # avoids self match
        self.assertRaises(DoorstopError, self.item.find_ref)

    def test_find_ref_none(self):
        """Verify nothing returned when no external reference is specified."""
        self.assertEqual((None, None), self.item.find_ref())

    def test_link_add(self):
        """Verify links can be added to an item."""
        self.item.add_link('abc')
        self.item.add_link('123')
        self.assertEqual(['123', 'abc'], self.item.links)

    def test_link_add_duplicate(self):
        """Verify duplicate links are ignored."""
        self.item.add_link('abc')
        self.item.add_link('abc')
        self.assertEqual(['abc'], self.item.links)

    def test_link_remove_duplicate(self):
        """Verify removing a link twice is not an error."""
        self.item.links = ['123', 'abc']
        self.item.remove_link('abc')
        self.item.remove_link('abc')
        self.assertEqual(['123'], self.item.links)

    def test_invalid_file_name(self):
        """Verify an invalid file name cannot be a requirement."""
        self.assertRaises(DoorstopError, MockItem, "path/to/REQ.yaml")
        self.assertRaises(DoorstopError, MockItem, "path/to/001.yaml")

    def test_invalid_file_ext(self):
        """Verify an invalid file extension cannot be a requirement."""
        self.assertRaises(DoorstopError, MockItem, "path/to/REQ001")
        self.assertRaises(DoorstopError, MockItem, "path/to/REQ001.txt")

    @patch('doorstop.core.item.Item', MockItem)
    def test_new(self):
        """Verify items can be created."""
        item = MockItem.new(EMPTY, FILES, 'TEST', 5, 42, (1, 2, 3))
        path = os.path.join(EMPTY, 'TEST00042.yml')
        self.assertEqual(path, item.path)
        self.assertEqual((1, 2, 3), item.level)
        MockItem._new.assert_called_once_with(path)

    def test_new_existing(self):
        """Verify an exception is raised if the item already exists."""
        self.assertRaises(DoorstopError,
                          Item.new, FILES, FILES, 'REQ', 3, 2, (1, 2, 3))

    def test_check_document(self):
        """Verify an item can be checked against a document."""
        document = Mock()
        document.parent = 'fake'
        self.item.check(document=document)

    def test_check_document_with_links(self):
        """Verify an item can be checked against a document with links."""
        self.item.add_link('unknown1')
        document = Mock()
        document.parent = 'fake'
        self.item.check(document=document)

    def test_check_tree(self):
        """Verify an item can be checked against a tree."""
        self.item.add_link('fake1')
        tree = Mock(find_item=Mock())
        self.item.check(tree=tree)

    def test_check_tree_error(self):
        """Verify an item can be checked against a tree with errors."""
        self.item.add_link('fake1')
        tree = Mock(find_item=Mock(side_effect=DoorstopError))
        self.assertRaises(DoorstopError, self.item.check, tree=tree)

    @patch('os.remove')
    def test_delete(self, mock_remove):
        """Verify an item can be deleted."""
        self.item.delete()
        mock_remove.assert_called_once_with(self.item.path)


class TestFormatting(unittest.TestCase):  # pylint: disable=R0904
    """Unit tests for text formatting in Items."""  # pylint: disable=C0103

    ITEM = os.path.join(FILES, 'REQ001.yml')

    def setUp(self):
        with open(self.ITEM, 'rb') as infile:
            self.backup = infile.read()

    def tearDown(self):
        with open(self.ITEM, 'wb') as outfile:
            outfile.write(self.backup)

    def test_load_save(self):
        """Verify text formatting is preserved."""
        item = Item(self.ITEM)
        item.load()
        item.save()
        with open(self.ITEM, 'rb') as infile:
            text = infile.read()
        self.assertEqual(self.backup, text)


class TestModule(unittest.TestCase):  # pylint: disable=R0904
    """Unit tests for the doorstop.core.item module."""  # pylint: disable=C0103

    def test_tbd(self):
        """Verify TBD."""
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



"""
Doorstop reporting functionality.
"""

import os
import textwrap
import markdown


def get_text(document, indent=8, width=79, ignored=None):
    """Yield lines for a text report.

    @param document: Document to publish
    @param indent: number of spaces to indent text
    @param width: maximum line length
    @param ignored: function to determine if a path should be skipped

    @return: iterator of lines of text
    """
    for item in document.items:

        level = '.'.join(str(l) for l in item.level)
        identifier = item.id

        # Level and ID
        yield "{l:<{s}}{i}".format(l=level, i=identifier, s=indent)

        # Text
        if item.text:
            yield ""  # break before text
            for line in item.text.splitlines():
                for chunk in _chunks(line, width, indent):
                    yield chunk

                if not line:  # pragma: no cover - integration test
                    yield ""  # break between paragraphs

        # Reference
        if item.ref:
            yield ""  # break before reference
            path, line = item.find_ref(ignored=ignored)
            relpath = os.path.relpath(path, item.root).replace('\\', '/')
            ref = "Reference: {p} (line {l})".format(p=relpath, l=line)
            for chunk in _chunks(ref, width, indent):
                yield chunk

        # Links
        if item.links:
            yield ""  # break before links
            links = "Links: " + ', '.join(item.links)
            for chunk in _chunks(links, width, indent):
                yield chunk

        yield ""  # break between items


def _chunks(text, width, indent):
    """Yield wrapped lines of text."""
    for chunk in textwrap.wrap(text, width,
                               initial_indent=' ' * indent,
                               subsequent_indent=' ' * indent):
        yield chunk


def get_markdown(document, ignored=None):
    """Yield lines for a Markdown report.

    @param document: Document to publish
    @param ignored: function to determine if a path should be skipped

    @return: iterator of lines of text
    """
    for item in document.items:

        heading = '#' * item.heading
        level = '.'.join(str(l) for l in item.level)
        identifier = item.id

        # Level and ID
        yield "{h} {l} ({i})".format(h=heading, l=level, i=identifier)

        # Text
        if item.text:
            yield ""  # break before text
            for line in item.text.splitlines():
                yield line

        # Reference
        if item.ref:
            yield ""  # break before reference
            path, line = item.find_ref(ignored=ignored)
            relpath = os.path.relpath(path, item.root).replace('\\', '/')
            ref = "Reference: {p} (line {l})".format(p=relpath, l=line)
            yield ref

        # Links
        if item.links:
            yield ""  # break before links
            links = '*' + "Links: " + ', '.join(item.links) + '*'
            yield links

        yield ""  # break between items


def get_html(document, ignored=None):
    """Yield lines for an HTML report.

    @param document: Document to publish
    @param ignored: function to determine if a path should be skipped

    @return: iterator of lines of text
    """
    lines = get_markdown(document, ignored=ignored)
    text = '\n'.join(lines)
    html = markdown.markdown(text)
    for line in html.splitlines():
        yield line



"""
Representation of items in a Doorstop document.
"""

import os
import re
import functools
import logging

import yaml


from doorstop.common import DoorstopError


# http://en.wikipedia.org/wiki/Sentence_boundary_disambiguation
SBD = re.compile(r"((?<=[a-z0-9][.?!])|(?<=[a-z0-9][.?!]\"))(\s|\r\n)(?=\"?[A-Z])")  # pylint: disable=C0301


class _literal(str):  # pylint: disable=R0904
    """Custom type for text which should be dumped in the literal style."""
    pass


def _literal_representer(dumper, data):
    """Return a custom dumper that formats str in the literal style."""
    style = '|' if data else ''
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style=style)

yaml.add_representer(_literal, _literal_representer)


def _auto_load(func):
    """Decorator for methods that should automatically load from file."""
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        """Wrapped method to call self.load() before execution."""
        self.load()
        return func(self, *args, **kwargs)
    return wrapped


def _auto_save(func):
    """Decorator for methods that should automatically save to file."""
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        """Wrapped method to call self.save() after execution."""
        result = func(self, *args, **kwargs)
        self.save()
        return result

    return wrapped


class Item(object):
    """Represents a file with linkable text that is part of a document."""

    EXTENSIONS = '.yml', '.yaml'

    DEFAULT_LEVEL = (1,)
    DEFAULT_TEXT = ""
    DEFAULT_REF = ""
    DEFAULT_LINKS = set()

    def __init__(self, path, root=os.getcwd(),
                 _level=None,
                 # TODO: are these arguments needed here?
                 _text=None, _ref=None, _links=None):
        """Load an item from an existing file.

        Internally, this constructor is also used to initialize new
        items by providing default properites.

        @param path: path to Item file
        @param root: path to root of project
        """
        # Check item's path
        if not os.path.isfile(path):
            raise DoorstopError("item does not exist: {}".format(path))
        # Check file name
        filename = os.path.basename(path)
        name, ext = os.path.splitext(filename)
        try:
            self.split_id(name)
        except DoorstopError:
            raise
        # Check file extension
        if ext.lower() not in self.EXTENSIONS:
            msg = "'{0}' extension not in {1}".format(path, self.EXTENSIONS)
            raise DoorstopError(msg)
        # Initialize Item
        self.path = path
        self.root = root
        self._exists = True
        self._level = _level or Item.DEFAULT_LEVEL
        self._text = _text or Item.DEFAULT_TEXT
        self._ref = _ref or Item.DEFAULT_REF
        self._links = _links or Item.DEFAULT_LINKS

    def __repr__(self):
        return "Item({})".format(repr(self.path))

    def __str__(self):
        relpath = os.path.relpath(self.path, self.root)
        return "{} (@{}{})".format(self.id, os.sep, relpath)

    def __eq__(self, other):
        return isinstance(other, Item) and self.path == other.path

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.level < other.level

    @staticmethod
    def new(path, root, prefix, digits, number, level):  # pylint: disable=R0913
        """Create a new item.

        @param path: path to directory for the new item
        @param root: path to root of the project
        @param prefix: prefix for the new item
        @param digits: number of digits for the new document
        @param number: number for the new item
        @param level: level for the new item (None for default)

        @raise DoorstopError: if the item already exists
        """
        identifier = Item.join_id(prefix, digits, number)
        filename = identifier + Item.EXTENSIONS[0]
        path2 = os.path.join(path, filename)
        # Check for an existing item
        if os.path.exists(path2):
            raise DoorstopError("item already exists: {}".format(path2))
        # Create the item file
        Item._new(path2)
        # Return the new item
        item = Item(path2, root=root, _level=level)
        item.save()
        return item

    @staticmethod
    def _new(path):  # pragma: no cover, integration test
        """Create a new item file.

        @param config: path to new item file
        """
        with open(path, 'w'):
            pass  # just touch the file

    def load(self, reload=False):
        """Load the item's properties from a file."""
        if getattr(self, '_loaded', False) and not reload:
            return
        logging.debug("loading {}...".format(repr(self)))
        # Read the YAML from file
        text = self._read()
        # Parse the YAML data
        try:
            data = yaml.load(text)
        except yaml.scanner.ScannerError as error:  # pylint: disable=E1101
            msg = "invalid contents: {}:\n{}".format(self, error)
            raise DoorstopError(msg)
        # Store parsed data
        if data:
            self._level = self._convert_level(data.get('level', self._level))
            self._text = data.get('text', self._text)
            self._ref = data.get('ref', self._ref)
            self._links = set(data.get('links', self._links))
        setattr(self, '_loaded', True)

    def _read(self):  # pragma: no cover, integration test
        """Read text from the file."""
        if not self._exists:
            raise DoorstopError("cannot load from deleted: {}".format(self))
        with open(self.path, 'rb') as infile:
            return infile.read().decode('UTF-8')

    def save(self):
        """Format and save the item's properties to a file."""
        logging.debug("saving {}...".format(repr(self)))
        # Collect the data items
        level = '.'.join(str(n) for n in self._level)
        if len(self._level) == 2:
            level = float(level)
        elif len(self._level) == 1:
            level = int(level)
        text = _literal(self._sbd(self._text))
        ref = self._ref.strip()
        links = sorted(self._links)
        # Build the data structure
        data = {'level': level,
                'text': text,
                'links': links,
                'ref': ref}
        # Dump the data to YAML
        dump = yaml.dump(data, default_flow_style=False)
        # Save the YAML to file
        self._write(dump)
        setattr(self, '_loaded', False)

    @staticmethod
    def _sbd(text):
        """Replace sentence boundaries with newlines.

        >>> Item._sbd("Hello, world!")
        'Hello, world!'

        >>> Item._sbd("Hello, world! How are you? I'm fine. Good.")
        "Hello, world!\\nHow are you?\\nI'm fine.\\nGood."

        """
        return SBD.sub('\n', text.strip())

    def _write(self, text):  # pragma: no cover, integration test
        """Write text to the file."""
        if not self._exists:
            raise DoorstopError("cannot save to deleted: {}".format(self))
        with open(self.path, 'wb') as outfile:
            outfile.write(bytes(text, 'UTF-8'))

    @property
    def id(self):  # pylint: disable=C0103
        """Get the item's ID."""
        return os.path.splitext(os.path.basename(self.path))[0]

    @staticmethod
    def join_id(prefix, digits, number):
        """Join the parts of an item's ID into an ID.

        >>> Item.join_id('ABC', 5, 123)
        'ABC00123'

        """
        return "{}{}".format(prefix, str(number).zfill(digits))

    @staticmethod
    def split_id(text):
        """Split an item's ID into a prefix and number.

        >>> Item.split_id('ABC00123')
        ('ABC', 123)

        """
        match = re.match(r"([a-zA-Z]+)(\d+)", text)
        if not match:
            raise DoorstopError("invalid ID: {}".format(text))
        return match.group(1), int(match.group(2))

    @property
    def prefix(self):
        """Get the item ID's prefix."""
        return self.split_id(self.id)[0]

    @property
    def number(self):
        """Get the item ID's number."""
        return self.split_id(self.id)[1]

    @property
    @_auto_load
    def level(self):
        """Get the item level."""
        return self._level

    @level.setter
    @_auto_save
    def level(self, level):
        """Set the item's level."""
        self._level = self._convert_level(level)

    @property
    def heading(self):
        """Get the heading order based on the level."""
        level = list(self.level)
        while level[-1] == 0:
            del level[-1]
        return len(level)

    @staticmethod
    def _convert_level(text):
        """Convert a level string to a tuple.

        >>> Item._convert_level("1.2.3")
        (1, 2, 3)

        >>> Item._convert_level(['4', '5'])
        (4, 5)

        >>> Item._convert_level(4.2)
        (4, 2)

        """
        # Correct for integers (42) and floats (4.2) in YAML
        if isinstance(text, int) or isinstance(text, float):
            text = str(text)
        # Split strings by periods
        if isinstance(text, str):
            nums = text.split('.')
        else:
            nums = text
        return tuple(int(n) for n in nums)

    @property
    @_auto_load
    def text(self):
        """Get the item's text."""
        return self._text

    @text.setter
    @_auto_save
    def text(self, text):
        """Set the item's text."""
        self._text = text

    @property
    @_auto_load
    def ref(self):
        """Get the item's external file reference."""
        return self._ref

    @ref.setter
    @_auto_save
    def ref(self, ref):
        """Set the item's external file reference."""
        self._ref = ref

    @_auto_load
    def find_ref(self, root=None, ignored=None):
        """Find the external file reference and line number.

        @param root: override the path to the working copy (for testing)
        @param ignored: function to determine if a path should be skipped

        @return: path to file, line number (or None, None when no ref)

        @raise DoorstopError: when no ref is found
        """
        if not self.ref:
            logging.debug("no external reference to search for")
            return None, None
        ignored = ignored or (lambda _: False)
        logging.info("seraching for ref '{}'...".format(self.ref))
        pattern = r"(\b|\W){}(\b|\W)".format(re.escape(self.ref))
        logging.debug("regex: {}".format(pattern))
        regex = re.compile(pattern)
        for root, _, filenames in os.walk(root or self.root):
            for filename in filenames:  # pragma: no cover, integration test
                path = os.path.join(root, filename)
                # Skip the item's file while searching
                if path == self.path:
                    continue
                # Skip hidden directories
                if (os.path.sep + '.') in path:
                    continue
                # Skip ignored paths
                if ignored(path):
                    continue
                # Search for the reference in the file
                try:
                    with open(path, 'r') as external:
                        for index, line in enumerate(external):
                            if regex.search(line):
                                logging.info("found ref: {}".format(path))
                                return path, index + 1
                except UnicodeDecodeError:
                    pass
        msg = "external reference not found: {}".format(self.ref)
        raise DoorstopError(msg)

    @property
    @_auto_load
    def links(self):
        """Get the items this item links to."""
        return sorted(self._links)

    @links.setter
    @_auto_save
    def links(self, links):
        """Set the items this item links to."""
        self._links = set(links)

    @_auto_load
    @_auto_save
    def add_link(self, item):
        """Add a new link to another item."""
        self._links.add(item)

    @_auto_load
    @_auto_save
    def remove_link(self, item):
        """Remove an existing link."""
        try:
            self._links.remove(item)
        except KeyError:
            logging.warning("link to {0} does not exist".format(item))

    def check(self, document=None, tree=None, ignored=None):
        """Confirm the item is valid.

        @param document: document to validate the item
        @param tree: tree to validate the item
        @param ignored: function to determine if a path should be skipped

        @return: indication that the item is valid
        """
        logging.info("checking item {}...".format(self))
        # Verify the file can be parsed
        self.load()
        # Check text
        if not self.text and not self.ref:
            logging.warning("no text: {}".format(self))
        # Check external references
        self.find_ref(ignored=ignored)
        # Check links against the document
        if document:
            if document.parent and self.level[-1] != 0 and not self.links:
                logging.warning("no links: {}".format(self))

            for identifier in self.links:
                prefix = self.split_id(identifier)[0]
                if prefix.lower() != document.parent.lower():
                    msg = "link to non-parent '{}' in {}".format(identifier,
                                                                 self)
                    logging.warning(msg)
        # Check links against the tree
        if tree:
            for identifier in self.links:
                item = tree.find_item(identifier)
                logging.debug("found linked item: {}".format(item))

        # Reformat the file
        self.save()
        # Item is valid
        return True

    def delete(self):
        """Delete the item from the file system."""
        logging.info("deleting {}...".format(self.path))
        os.remove(self.path)
        self._exists = False  # prevent furthur access



#!/usr/bin/env python

"""
Unit tests for the doorstop.core.report module.
"""

import unittest
from unittest.mock import Mock

import os

from doorstop.core.report import get_text, get_markdown, get_html
from doorstop.core.vcs.mockvcs import WorkingCopy

from doorstop.core.test import FILES
from doorstop.core.test.test_item import MockItem

ASSERT_CONTENTS = True  # set to False to overrite contents


class TestModule(unittest.TestCase):  # pylint: disable=R0904
    """Unit tests for the doorstop.core.report module."""  # pylint: disable=C0103

    @classmethod
    def setUpClass(cls):
        cls.document = Mock()
        cls.document.items = [
            MockItem('path/to/req3.yml',
                     _file="links: [sys3]\ntext: '" + ("Hello, world! " * 10)
                     + "'\nlevel: 1.0"),
            MockItem('path/to/req1.yml',
                     _file="links: []\ntext: 'abc\n123'\nlevel: 1.1"),
            MockItem('path/to/req2.yml',
                     _file="links: [sys1, sys2]\ntext: ''\nlevel: 2"),
            MockItem('path/to/req4.yml',
                     _file="links: [sys2]\nref: 'r1'\nlevel: 2.1.1"),
        ]
        cls.work = WorkingCopy(None)

    def test_get_text(self):
        """Verify a text report can be created."""
        path = os.path.join(FILES, 'report.txt')
        expected = open(path).read()
        lines = get_text(self.document, ignored=self.work.ignored)
        text = ''.join(line + '\n' for line in lines)
        if ASSERT_CONTENTS:
            self.assertEqual(expected, text)
        with open(path, 'wb') as outfile:
            outfile.write(bytes(text, 'utf-8'))

    def test_get_markdown(self):
        """Verify a Markdown report can be created."""
        path = os.path.join(FILES, 'report.md')
        expected = open(path).read()
        lines = get_markdown(self.document, ignored=self.work.ignored)
        text = ''.join(line + '\n' for line in lines)
        if ASSERT_CONTENTS:
            self.assertEqual(expected, text)
        with open(path, 'wb') as outfile:
            outfile.write(bytes(text, 'utf-8'))

    def test_get_html(self):
        """Verify an HTML report can be created."""
        path = os.path.join(FILES, 'report.html')
        expected = open(path).read()
        lines = get_html(self.document, ignored=self.work.ignored)
        text = ''.join(line + '\n' for line in lines)
        if ASSERT_CONTENTS:
            self.assertEqual(expected, text)
        with open(path, 'wb') as outfile:
            outfile.write(bytes(text, 'utf-8'))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""
Unit tests for the doorstop.vcs.base module.
"""

import unittest


class TestModule(unittest.TestCase):  # pylint: disable=R0904
    """Tests for the doorstop.vcs.base module."""  # pylint: disable=C0103

    def test_tbd(self):
        """Verify TBD."""
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()
